import argparse
from os.path import join
from multiprocessing import Pool, cpu_count

from ctc_metrics.metrics import valid
from ctc_metrics.utils.handle_results import print_results
from ctc_metrics.utils.filesystem import \
    parse_directories, read_tracking_file, parse_masks
from ctc_metrics.utils.representations import match as match_tracks


def validate_sequence(
        res: str,
        threads: int = 0,
):
    """
    Validates a single sequence

    Args:
        res: The path to the results.
        threads: The number of threads to use. If 0, the number of threads
            is set to the number of available CPUs.

    Returns:
        The results stored in a dictionary.
    """
    print("\r", res, end="")
    res_tracks = read_tracking_file(join(res, "res_track.txt"))
    res_masks = parse_masks(res)
    assert len(res_masks) > 0, res
    args = zip([None for x in res_masks], res_masks)
    if threads != 1:
        if threads == 0:
            threads = cpu_count()
        with Pool(threads) as p:
            matches = p.starmap(match_tracks, args)
    else:
        matches = [match_tracks(*x) for x in args]
    labels_gt, labels_res, mapped_gt, mapped_res = [], [], [], []
    for match in matches:
        labels_gt.append(match[0])
        labels_res.append(match[1])
        mapped_gt.append(match[2])
        mapped_res.append(match[3])
    results = {"Valid": valid(res_masks, res_tracks, labels_res)}
    print("\r", end="")
    return results


def validate_all(
        res_root: str,
        threads: int = 0,
):
    """
    Evaluate all sequences in a directory

    Args:
        res_root: The path to the result directory.
        threads: The number of threads to use. If 0, the number of threads
            is set to the number of available CPUs.

    Returns:
        The results stored in a dictionary.
    """
    results = []
    ret = parse_directories(res_root, None)
    for res, _, name in zip(*ret):
        results.append([name, validate_sequence(res, threads)])
    return results


def parse_args():
    """ Parses the arguments. """
    parser = argparse.ArgumentParser(description='Validates CTC-Sequences.')
    parser.add_argument('--res', type=str, required=True)
    parser.add_argument('-r', '--recursive', action="store_true")
    parser.add_argument('-n', '--num-threads', type=int, default=0)
    args = parser.parse_args()
    return args


def main():
    """
    Main function that is called when the script is executed.
    """
    args = parse_args()
    if args.recursive:
        res = validate_all(args.res, args.num_threads)
    else:
        res = validate_sequence(args.res, args.num_threads)
    print_results(res)


if __name__ == "__main__":
    main()




import numpy as np

from ctc_metrics.metrics.biological.bc import get_ids_that_ends_with_split


def is_valid_track(
        tracks: np.ndarray,
):
    """
    Extracts the boolean indices of valid tracks that start and end with a
        cell split.

    Args:
        tracks: The tracks to check.

    Returns:
        The boolean indices of valid tracks that start and end with a cell
            split.
    """
    ends_with_split = get_ids_that_ends_with_split(tracks)
    is_parent = np.isin(tracks[:, 0], ends_with_split)
    is_child = np.isin(tracks[:, 3], ends_with_split)
    valid = np.logical_and(is_parent, is_child)
    return valid


def cca(
        comp_tracks: np.ndarray,
        ref_tracks: np.ndarray
):
    """
    Computes the cell cycle accuracy. As described in the paper,
         "An objective comparison of cell-tracking algorithms."
           - Vladimir Ulman et al., Nature methods 2017

    Args:
        comp_tracks: The result tracks. A (n,4) numpy ndarray with columns:
            - label
            - birth frame
            - end frame
            - parent
        ref_tracks: The ground truth tracks. A (n,4) numpy ndarray with columns:
            - label
            - birth frame
            - end frame
            - parent

    Returns:
        The cell cycle accuracy metric.
    """

    # Extract relevant tracks with parents and children in reference
    valid_ref = is_valid_track(ref_tracks)
    if np.sum(valid_ref) == 0:
        return None
    track_lengths_ref = ref_tracks[valid_ref, 2] - ref_tracks[valid_ref, 1]
    # Extract relevant tracks with parents and children in computed result
    valid_comp = is_valid_track(comp_tracks)
    if np.sum(valid_comp) == 0:
        return 0
    track_lengths_comp = comp_tracks[valid_comp, 2] - comp_tracks[valid_comp, 1]
    # Calculate CCA
    max_track_length = np.max(
        [np.max(track_lengths_ref), np.max(track_lengths_comp)])
    hist_ref = np.zeros(np.max(max_track_length) + 1)
    for i in track_lengths_ref:
        hist_ref[i] += 1
    hist_ref = hist_ref / np.sum(hist_ref)
    cum_hist_ref = np.cumsum(hist_ref)
    hist_comp = np.zeros(np.max(max_track_length) + 1)
    for i in track_lengths_comp:
        hist_comp[i] += 1
    hist_comp = hist_comp / np.sum(hist_comp)
    cum_hist_comp = np.cumsum(hist_comp)
    return float(1 - np.max(np.abs(cum_hist_ref - cum_hist_comp)))




def op_csb(
        seg: float,
        det: float,
):
    """
    Computes the OP_CSB metric. As described by
        [celltrackingchallenge](http://celltrackingchallenge.net/).

    Args:
        seg: The segmentation metric.
        det: The detection metric.

    Returns:
        The OP_CSB metric.
    """

    op = 0.5 * seg + 0.5 * det
    return op



from test.utils import test_seq_res
from ctc_metrics import validate_sequence


def test_validate_sequence():
    res = validate_sequence(test_seq_res)
    assert bool(res["Valid"]) is True, f"{bool(res['Valid'])} != True"



import os.path

import numpy as np
import tifffile as tiff
from sklearn.metrics import confusion_matrix
from scipy.sparse import lil_array


def track_confusion_matrix(
        labels_ref: list,
        labels_comp: list,
        mapped_ref: list,
        mapped_comp: list
):
    """
    Computes the confusion matrix for the input data.

    Args:
        labels_ref: The labels of the ground truth masks.
        labels_comp: The labels of the result masks.
        mapped_ref: The matched labels of the ground truth masks.
        mapped_comp: The matched labels of the result masks.

    Returns:
        The confusion matrix. The size if the confusion matrix is
            max_label_ref + 1 x max_label_comp + 1
        where [:, 0] contains the false negatives and [0, :] contains the
        false positives. The rest of the matrix contains the true positives.

    """
    max_label_ref = int(np.max(np.concatenate(labels_ref)))
    max_label_comp = int(np.max(np.concatenate(labels_comp)))

    # Gather association data
    track_intersection = np.zeros((max_label_ref + 1, max_label_comp + 1))

    for ref, comp, m_ref, m_comp in zip(
            labels_ref, labels_comp, mapped_ref, mapped_comp):
        # Fill track intersection matrix
        ref = np.asarray(ref)
        comp = np.asarray(comp).astype(int)
        m_ref = np.asarray(m_ref).astype(int)
        m_comp = np.asarray(m_comp).astype(int)
        if len(m_ref) > 0:
            track_intersection[m_ref, m_comp] += 1
        fna = ref[np.isin(ref, m_ref, invert=True)]
        track_intersection[fna, 0] += 1
        fpa = comp[np.isin(comp, m_comp, invert=True)].astype(int)
        track_intersection[0, fpa] += 1

    return track_intersection


def match(
        ref_path: str,
        comp_path: str
):
    """
    Matches the labels of the masks from the reference and computed result path.
    A label is matched if the intersection of a computed and reference mask is
    greater than 50% of the area of the reference mask.

    Args:
        ref_path: Path to the reference mask.
        comp_path: Path to the computed mask.

    Returns:
        A tuple of five numpy arrays. The first array contains the existing
        labels in the reference mask. The second array contains the
        existing labels in the computed mask. The third array contains the
        matched labels in the referenced mask. The fourth array contains
        the corresponding matched labels in the computed mask. The fifth
        array contains the intersection over union (IoU) for each matched
        label pair.
    """
    # Read the input data
    if comp_path is None:
        map_ref = tiff.imread(ref_path)
        labels_ref = np.unique(map_ref)
        labels_ref = labels_ref[labels_ref != 0]
        return labels_ref.tolist(), [], [], [], []
    if ref_path is None:
        # For trivial cases where only one mask should be analysed
        ref_path = comp_path
    if os.path.exists(ref_path):
        # No slices
        map_ref = tiff.imread(ref_path)
        map_com = tiff.imread(comp_path)
    else:
        # Slices in 3D dataset segmentation data
        dirname, base = os.path.dirname(ref_path), os.path.basename(ref_path)
        base = base.replace(".tif", "")
        slice_files = [x for x in os.listdir(dirname) if
                 x.endswith(".tif") and x.startswith(base)]
        slice_files = sorted(slice_files)
        map_ref = [tiff.imread(os.path.join(dirname, x)) for x in slice_files]
        map_ref = np.stack(map_ref, axis=0)
        slices = [x.replace(base+"_", "").replace(".tif", "")
                  for x in slice_files]
        slices = [int(x) for x in slices]
        map_com = tiff.imread(comp_path)
        _map_com = [map_com[x] for x in slices]
        map_com = np.stack(_map_com, axis=0)
    # Get the labels of the two masks (including background label 0)
    labels_ref, labels_comp = np.unique(map_ref), np.unique(map_com)
    if ref_path == comp_path:
        # For trivial cases where only one mask should be analysed
        iou = np.ones(len(labels_ref))
        labels_ref = labels_ref[labels_ref != 0]
        labels_comp = labels_comp[labels_comp != 0]
        return labels_ref.tolist(), labels_comp.tolist(), labels_ref.tolist(),\
            labels_comp.tolist(), iou.tolist()
    # Add offset to separate the labels of the two masks
    offset = int(np.max(labels_ref) + 1)
    map_com += offset
    # Compute the confusion matrix
    cm = confusion_matrix(map_ref.flatten(), map_com.flatten())
    sum_ref = np.sum(cm, axis=1, keepdims=True)
    sum_comp = np.sum(cm, axis=0, keepdims=True)
    # Compute the intersection over reference
    intersection_over_ref = cm / np.maximum(sum_ref, 1)
    # Compute the intersection over union (relevant to calculate SEG)
    intersection_over_union = cm / np.maximum(sum_ref + sum_comp - cm, 1)
    # Remove the background label and redundant parts of the matrix
    intersection_over_ref = \
        intersection_over_ref[1:len(labels_ref), 1 + len(labels_ref):]
    intersection_over_union = \
        intersection_over_union[1:len(labels_ref), 1 + len(labels_ref):]
    # Find matches according to AOGM (min 50% of ref needs to be covered)
    intersection_over_ref[intersection_over_ref <= 0.5] = 0
    intersection_over_ref[intersection_over_ref > 0.5] = 1
    # Create mapping between reference and computed labels
    rows, cols = np.nonzero(intersection_over_ref)
    labels_ref = labels_ref[1:]
    labels_comp = labels_comp[1:]
    mapped_ref = labels_ref[rows].tolist()
    mapped_comp = labels_comp[cols].tolist()
    iou = intersection_over_union[rows, cols].tolist()
    assert np.unique(mapped_ref).size == len(mapped_ref), \
        f"Reference node assigned to multiple computed nodes! " \
        f"{ref_path} {labels_ref, labels_comp, mapped_ref, mapped_comp}"
    labels_ref = labels_ref.tolist()
    labels_comp = labels_comp.tolist()
    return labels_ref, labels_comp, mapped_ref, mapped_comp, iou


def create_edge_mapping(
        tracks: np.ndarray,
        labels: list,
        V_tp: np.ndarray,
        cum_inds: np.ndarray,
):
    """
    Creates the edge mapping for the input tracks. The edge mapping is a
    nd.array with the following style:
        [[ind1, id1, det_test1, t1, ind2, id2, det_test2, t2, semantic]]
    where ind1 and ind2 are the indices of the nodes in V_tp, id1 and
    id2 are the labels of the vertices, det_test1 and det_test2 are the
    detection test of the vertices, t1 and t2 are the time steps of the
    vertices and semantic is the semantic label of the edge (0 for track link,
    1 for parent link).

    Args:
        tracks: The tracks. If None, no edges will be created
        labels: The labels of the ground truth masks.
        V_tp: The detection test matrix.
        cum_inds: The cumulative indices of the vertices per frame.

    Returns:
        The edge mapping.
    """
    if tracks is None:
        return np.zeros((0, 9))
    all_edges = []
    # Add track links
    ind_v = 0
    current_t = 0
    for l_gt1, l_gt2 in zip(labels[:-1], labels[1:]):
        l_gt1, l_gt2 = np.array(l_gt1), np.array(l_gt2)
        mapping = l_gt1[:, None] == l_gt2[None, :]
        ind1, ind2 = np.where(mapping)
        id1, id2 = l_gt1[ind1], l_gt2[ind2]
        t1, t2 = np.ones_like(id1) * current_t, np.ones_like(id1) * current_t + 1
        ind1, ind2 = ind1 + ind_v, ind2 + ind_v + len(l_gt1)
        edges = np.stack([
            ind1, id1, V_tp[ind1], t1,
            ind2, id2,  V_tp[ind2], t2,
            np.zeros_like(id1)], axis=1)
        all_edges.append(edges)
        ind_v += len(l_gt1)
        current_t += 1
    # Add track and parent links from track file
    for track in tracks:
        label2, birth2, _, parent2 = track
        if parent2 == 0:
            continue
        label1, _, end1, _ = tracks[tracks[:, 0] == parent2][0]
        ind1 = np.argwhere(labels[end1] == label1)[0] + cum_inds[end1]
        ind2 = np.argwhere(labels[birth2] == label2)[0] + cum_inds[birth2]
        ind1, ind2 = ind1.item(), ind2.item()
        edges = np.asarray(
            [ind1, label1, int(V_tp[ind1]), end1, ind2, label2, int(V_tp[ind2]), birth2, 1]
        )[None, :]
        all_edges.append(edges)
    return np.concatenate(all_edges, axis=0).astype(int)


def create_detection_test_matrix(
        num_V_C: int,
        num_V_R: int,
        labels_ref: list,
        labels_comp: list,
        mapped_ref: list,
        mapped_comp: list,
):
    """
    Creates the detection test matrix for the input tracks. The detection test
    is stored as a sparse matrix with num_V_C rows and num_V_R columns. The
    matrix is filled with 1 if the vertex is a match and 0 otherwise.

    Args:
        num_V_C: The number of vertices in the computed graph.
        num_V_R: The number of vertices in the reference graph.
        labels_ref: The labels of the ground truth masks.
        labels_comp: The labels of the result masks.
        mapped_ref: The matched labels of the ground truth masks.
        mapped_comp: The matched labels of the result masks.

    Returns:
        The detection test matrix.
    """
    det_test = lil_array((num_V_C, num_V_R))
    ind_v_r = 0
    ind_v_c = 0
    for l_gt, l_res, m_gt, m_res in zip(
            labels_ref, labels_comp, mapped_ref, mapped_comp
    ):
        i_r = np.searchsorted(l_gt, m_gt, sorter=np.argsort(l_gt)) + ind_v_r
        i_c = np.searchsorted(l_res, m_res, sorter=np.argsort(l_res)) + ind_v_c
        det_test[i_c, i_r] = 1
        ind_v_r += len(l_gt)
        ind_v_c += len(l_res)
    return det_test


def count_acyclic_graph_correction_operations(
        ref_tracks: np.ndarray,
        comp_tracks: np.ndarray,
        labels_ref: list,
        labels_comp: list,
        mapped_ref: list,
        mapped_comp: list,
):
    """
    Calculates the necessary operations to correct the result tracks to match
    the ground truth tracks. The operations are counted according to:
        Cell Tracking Accuracy Measurement Based on Comparison of Acyclic
        Oriented Graphs; Matula etal. 2015

    Args:
        comp_tracks: The result tracks.
        ref_tracks: The ground truth tracks.
        labels_ref: The labels of the ground truth masks.
        labels_comp: The labels of the result masks.
        mapped_ref: The matched labels of the ground truth masks.
        mapped_comp: The matched labels of the result masks.

    Returns:
        NS, FN, FP, ED, EA, EC, num_vertices, num_edges
    """
    # Count vertices in the input data
    stats = {}
    stats["num_vertices_R"] = np.sum([len(l) for l in labels_ref])
    stats["num_vertices_C"] = np.sum([len(l) for l in labels_comp])
    stats["num_vertices"] = stats["num_vertices_R"]
    # Cumulate the number of vertices per frame
    cum_inds_R = np.cumsum([0] + [len(l) for l in labels_ref])
    cum_inds_C = np.cumsum([0] + [len(l) for l in labels_comp])
    # Perform "detection test"
    det_test = create_detection_test_matrix(
        stats["num_vertices_C"], stats["num_vertices_R"],
        labels_ref, labels_comp, mapped_ref, mapped_comp
    )
    # Classify vertices to tp, fp, fn and vs
    assignments_r = np.sum(det_test, axis=0)
    assignments_c = np.sum(det_test, axis=1)
    assert np.max(assignments_r) <= 1
    V_tp_r = assignments_r == 1
    V_tp_c = assignments_c == 1
    stats["TP"] = np.sum(V_tp_r)
    stats["FN"] = np.sum(~V_tp_r)
    stats["FP"] = np.sum(assignments_c == 0)
    stats["VS"] = np.sum(assignments_c > 1)
    stats["NS"] = stats["TP"] - (np.sum(V_tp_c) + stats["VS"])
    # Mapping from reference to computed
    det_test[~V_tp_c, :] = 0
    comp, ref = det_test.nonzero()
    assert len(comp) == np.sum(V_tp_c)
    comp_to_ref = np.zeros(stats["num_vertices_C"]) * np.nan
    comp_to_ref[comp] = ref
    assert np.all(np.sort(comp) == comp)
    # Create edge mapping ...
    # ... for reference
    E_R = create_edge_mapping(ref_tracks, labels_ref, V_tp_r, cum_inds_R)
    # ... for computed
    E_C = create_edge_mapping(comp_tracks, labels_comp, V_tp_c, cum_inds_C)
    # Stop calculation if no edges are present in the computed graph (e.g. if
    #   only segmentation is present)
    if E_C.size == 0:
        stats["ED"] = 0
        stats["EA"] = len(E_R)
        stats["EC"] = 0
        stats["num_edges"] = len(E_R)
        return stats
    # Reduce the computed graph to an induced subgraph with only uniquely
    #   matched vertices
    E_C = E_C[(E_C[:, 2] * E_C[:, 6]) == 1]
    # Add mapping to Reference graph such that E_C is:
    #    ind1,id1,det_tst1,t1,ind2,id2,det_test2,t2,sem,ind1_R,ind2_R
    E_C = np.concatenate([
        E_C,
        comp_to_ref[E_C[:, 0]][:, None].astype(int),
        comp_to_ref[E_C[:, 4]][:, None].astype(int)
    ], axis=1)
    assert not np.any(np.isnan(E_C))
    # Map the edges to edges
    unique_edge_ids_R = (E_R[:, 0] * 10 ** len(str(stats["num_vertices_R"]))
                         + E_R[:, 4])
    unique_edge_ids_C = (E_C[:, 9] * 10 ** len(str(stats["num_vertices_R"]))
                         + E_C[:, 10])
    if unique_edge_ids_R.size > 0:
        assert np.max(np.unique(unique_edge_ids_R, return_counts=True)[1]) == 1
    if unique_edge_ids_C.size > 0:
        assert np.max(np.unique(unique_edge_ids_C, return_counts=True)[1]) == 1
    isin_R = np.isin(unique_edge_ids_C, unique_edge_ids_R)
    isin_C = np.isin(unique_edge_ids_R, unique_edge_ids_C)
    E_R_mapped = E_R[isin_C]
    E_C_mapped = E_C[isin_R]
    E_R_mapped = E_R_mapped[np.argsort(unique_edge_ids_R[isin_C])]
    E_C_mapped = E_C_mapped[np.argsort(unique_edge_ids_C[isin_R])]
    # Calculate relevant edge statistics
    stats["ED"] = len(E_C[~isin_R])
    stats["EA"] = np.sum(~isin_C)
    stats["EC"] = len(E_C_mapped[E_C_mapped[:, 8] != E_R_mapped[:, 8]])
    stats["num_edges"] = len(E_R)
    return stats


def assign_comp_to_ref(
        labels_ref: list,
        mapped_ref: list,
        mapped_comp: list,
):
    """
    Assigns the computed labels to the reference labels.

    Args:
        labels_ref: The labels of the ground truth masks.
        mapped_ref: The matched labels of the ground truth masks.
        mapped_comp: The matched labels of the result masks.

    Returns:
        A dictionary with the reference labels as keys and the computed
        labels as values.
    """
    all_labels = np.unique(np.concatenate(labels_ref))
    max_frame = len(labels_ref)
    track_assignments = {
        k: np.zeros(max_frame) * np.nan for k in all_labels
    }
    frame = 0
    for l_gt, m_gt, m_res in zip(
            labels_ref, mapped_ref, mapped_comp
    ):
        for i in l_gt:
            if i in m_gt:
                m = m_res[np.argwhere(np.asarray(m_gt) == i)[0].item()]
                track_assignments[i][frame] = m
                counts = np.sum(np.asarray(m_res) == m)
                if counts > 1:
                    track_assignments[i][frame] = 0
            else:
                track_assignments[i][frame] = 0
        frame += 1
    return track_assignments


def merge_tracks(
        tracks: np.ndarray,
        labels: list,
        mapped: list
):
    """
    Merges tracks that belong to the same cell trajectory. This can happen, if
    a cell is invisible for a few frames and the tracking algorithm splits the
    track into two or more tracks. The resulting data is a relabelled version
    of the input data.

    Args:
        tracks: The tracks.
        labels: The labels of the masks.
        mapped: The matched labels of the masks.

    Returns:
        As input data, but relabelled.
    """
    tracks = np.copy(tracks)
    labels = [np.copy(x) for x in labels]
    mapped = [np.copy(x) for x in mapped]
    # Find tracks that belong together
    parents, cnts = np.unique(tracks[tracks[:, 3] > 0, 3], return_counts=True)
    parents = parents[cnts == 1].tolist()
    children = [tracks[tracks[:, 3] == p][0, 0] for p in parents]

    # Merge tracks
    mapping = {x: x for x in np.unique(tracks[:, 0])}
    for parent, child in zip(parents, children):
        for k, v in mapping.items():
            if v == child:
                mapping[k] = mapping[parent]

    # Relabel such that the labels are continuous
    remaining_labels = sorted(np.unique(list(mapping.values())))
    new_labels = list(range(1, len(remaining_labels) + 1))
    for k, v in mapping.items():
        new_v = new_labels[remaining_labels.index(v)]
        mapping[k] = new_v

    lut = np.zeros(np.max(list(mapping.keys())) + 1, dtype=int)
    for k, v in mapping.items():
        lut[k] = v

    # Relabel tracks
    new_tracks = np.copy(tracks)
    for k, v in mapping.items():
        new_tracks[tracks[:, 0] == k, 0] = v
        new_tracks[tracks[:, 3] == k, 3] = v
    ids, counts = np.unique(new_tracks[:, 0], return_counts=True)
    ids = ids[counts > 1]
    for i in ids:
        inds = np.argwhere(new_tracks[:, 0] == i).flatten()
        start = np.min(new_tracks[inds, 1])
        end = np.max(new_tracks[inds, 2])
        new_tracks[inds, 1] = start
        new_tracks[inds, 2] = end
        new_tracks = np.delete(new_tracks, inds[1:], axis=0)

    # Relabel labels and mapped
    new_labels = [[lut[x] for x in i] for i in labels]
    new_mapped = [[lut[x] for x in i] for i in mapped]

    return new_tracks, new_labels, new_mapped







import argparse
from os.path import join, basename
from multiprocessing import Pool, cpu_count
import numpy as np

from ctc_metrics.metrics import (
    valid, det, seg, tra, ct, tf, bc, cca, mota, hota, idf1, chota, mtml, faf,
    op_ctb, op_csb, bio, op_clb, lnk
)
from ctc_metrics.metrics import ALL_METRICS
from ctc_metrics.utils.handle_results import print_results, store_results
from ctc_metrics.utils.filesystem import parse_directories, read_tracking_file,\
    parse_masks
from ctc_metrics.utils.representations import match as match_tracks, \
    count_acyclic_graph_correction_operations, merge_tracks


def match_computed_to_reference_masks(
        ref_masks: list,
        comp_masks: list,
        threads: int = 0,
):
    """
    Matches computed masks to reference masks.

    Args:
        ref_masks: The reference masks. A list of paths to the reference masks.
        comp_masks: The computed masks. A list of paths to the computed masks.
        threads: The number of threads to use. If 0, the number of threads
            is set to the number of available CPUs.

    Returns:
        The results stored in a dictionary. The dictionary contains the
        following keys:
            - labels_ref: The reference labels. A list of lists containing
                the labels of the reference masks.
            - labels_comp: The computed labels. A list of lists containing
                the labels of the computed masks.
            - mapped_ref: The mapped reference labels. A list of lists
                containing the mapped labels of the reference masks.
            - mapped_comp: The mapped computed labels. A list of lists
                containing the mapped labels of the computed masks.
            - ious: The intersection over union values. A list of lists
                containing the intersection over union values between mapped
                reference and computed masks.
    """
    labels_ref, labels_comp, mapped_ref, mapped_comp, ious = [], [], [], [], []
    if threads != 1:
        if threads == 0:
            threads = cpu_count()
        with Pool(threads) as p:
            matches = p.starmap(match_tracks, zip(ref_masks, comp_masks))
    else:
        matches = [match_tracks(*x) for x in zip(ref_masks, comp_masks)]
    for match in matches:
        labels_ref.append(match[0])
        labels_comp.append(match[1])
        mapped_ref.append(match[2])
        mapped_comp.append(match[3])
        ious.append(match[4])
    return {
        "labels_ref": labels_ref,
        "labels_comp": labels_comp,
        "mapped_ref": mapped_ref,
        "mapped_comp": mapped_comp,
        "ious": ious
    }


def load_data(
        res: str,
        gt: str,
        trajectory_data: True,
        segmentation_data: True,
        threads: int = 0,
):
    """
    Load data that is necessary to calculate metrics from the given directories.

    Args:
        res: The path to the results.
        gt: The path to the ground truth.
        trajectory_data: A flag if trajectory data is available.
        segmentation_data: A flag if segmentation data is available.
        threads: The number of threads to use. If 0, the number of threads
            is set to the number of available CPUs.

    Returns:
        The computed tracks, the reference tracks, the trajectory data, the
        segmentation data, the computed masks and a flag if the results are
        valid.

    """
    # Read tracking files and parse mask files
    comp_tracks = read_tracking_file(join(res, "res_track.txt"))
    ref_tracks = read_tracking_file(join(gt, "TRA", "man_track.txt"))
    comp_masks = parse_masks(res)
    ref_tra_masks = parse_masks(join(gt, "TRA"))
    assert len(ref_tra_masks) > 0, f"{gt}: Ground truth masks is 0!)"
    assert len(ref_tra_masks) == len(comp_masks), (
        f"{res}: Number of result masks ({len(comp_masks)}) unequal to "
        f"the number of ground truth masks ({len(ref_tra_masks)})!)")
    # Match golden truth tracking masks to result masks
    traj = {}
    is_valid = 1
    if trajectory_data:
        traj = match_computed_to_reference_masks(
            ref_tra_masks, comp_masks, threads=threads)
        is_valid = valid(comp_masks, comp_tracks, traj["labels_comp"])
    # Match golden truth segmentation masks to result masks
    segm = {}
    if segmentation_data:
        ref_seg_masks = parse_masks(join(gt, "SEG"))
        _res_masks = [
            comp_masks[int(basename(x).replace(
                "man_seg", "").replace(".tif", "").replace("_", ""))]
            for x in ref_seg_masks
        ]
        segm = match_computed_to_reference_masks(
            ref_seg_masks, _res_masks, threads=threads)
    return comp_tracks, ref_tracks, traj, segm, comp_masks, is_valid


def calculate_metrics(
        comp_tracks: np.ndarray,
        ref_tracks: np.ndarray,
        traj: dict,
        segm: dict,
        metrics: list = None,
        is_valid: bool = None,
):  # pylint: disable=too-complex
    """
    Calculate metrics for given data.

    Args:
        comp_tracks: The computed tracks.A (n,4) numpy ndarray with columns:
            - label
            - birth frame
            - end frame
            - parent
        ref_tracks: The reference tracks. A (n,4) numpy ndarray with columns:
            - label
            - birth frame
            - end frame
            - parent
        traj: The frame-wise trajectory match data.
        segm: The frame-wise segmentation match data.
        metrics: The metrics to evaluate.
        is_valid: A Flag if the results are valid

    Returns:
        The results stored in a dictionary.
    """
    # Check if results are valid
    results = {x: None for x in metrics}
    if not is_valid:
        print("Invalid results!")
        results["Valid"] = 0
        return results

    # Create merge tracks
    if traj:
        new_tracks, new_labels, new_mapped = merge_tracks(
            ref_tracks, traj["labels_ref"], traj["mapped_ref"])
        traj["ref_tracks_merged"] = new_tracks
        traj["labels_ref_merged"] = new_labels
        traj["mapped_ref_merged"] = new_mapped
        new_tracks, new_labels, new_mapped = merge_tracks(
            comp_tracks, traj["labels_comp"], traj["mapped_comp"])
        traj["comp_tracks_merged"] = new_tracks
        traj["labels_comp_merged"] = new_labels
        traj["mapped_comp_merged"] = new_mapped

    # Prepare intermediate results
    graph_operations = {}
    if "DET" in metrics or "TRA" in metrics:
        graph_operations = \
            count_acyclic_graph_correction_operations(
                ref_tracks, comp_tracks,
                traj["labels_ref"], traj["labels_comp"],
                traj["mapped_ref"], traj["mapped_comp"]
            )

    # Calculate metrics
    if "Valid" in metrics:
        results["Valid"] = is_valid

    if "CHOTA" in metrics:
        results.update(chota(
            traj["ref_tracks_merged"], traj["comp_tracks_merged"],
            traj["labels_ref_merged"], traj["labels_comp_merged"],
            traj["mapped_ref_merged"], traj["mapped_comp_merged"]))

    if "DET" in metrics:
        results["DET"] = det(**graph_operations)

    if "SEG" in metrics:
        results["SEG"] = seg(segm["labels_ref"], segm["ious"])

    if "TRA" in metrics:
        _tra, _aogm, _aogm0 = tra(**graph_operations)
        results["TRA"] = _tra
        results["AOGM"] = _aogm
        results["AOGM_0"] = _aogm0
        for key in ("NS", "FN", "FP", "ED", "EA", "EC"):
            results[f"AOGM_{key}"] = graph_operations[key]

    if "LNK" in metrics:
        results["LNK"] = lnk(**graph_operations)

    if "DET" in metrics and "SEG" in metrics:
        results["OP_CSB"] = op_csb(results["SEG"], results["DET"])

    if "SEG" in metrics and "TRA" in metrics:
        results["OP_CTB"] = op_ctb(results["SEG"], results["TRA"])

    if "CT" in metrics:
        results["CT"] = ct(
            comp_tracks, ref_tracks,
            traj["labels_ref"], traj["mapped_ref"], traj["mapped_comp"])

    if "TF" in metrics:
        results["TF"] = tf(
            ref_tracks,
            traj["labels_ref"], traj["mapped_ref"], traj["mapped_comp"])

    if "BC" in metrics:
        for i in range(4):
            results[f"BC({i})"] = bc(
                comp_tracks, ref_tracks,
                traj["mapped_ref"], traj["mapped_comp"],
                i=i)

    if "CCA" in metrics:
        results["CCA"] = cca(comp_tracks, ref_tracks)

    if "CT" in metrics and "BC" in metrics and \
            "CCA" in metrics and "TF" in metrics:
        for i in range(4):
            results[f"BIO({i})"] = bio(
                results["CT"], results["TF"],
                results[f"BC({i})"], results["CCA"])

    if "BIO" in results and "LNK" in results:
        for i in range(4):
            results[f"OP_CLB({i})"] = op_clb(
                results["LNK"], results[f"BIO({i})"])

    if "MOTA" in metrics:
        results.update(mota(
            traj["labels_ref_merged"], traj["labels_comp_merged"],
            traj["mapped_ref_merged"], traj["mapped_comp_merged"]))

    if "HOTA" in metrics:
        results.update(hota(
            traj["labels_ref_merged"], traj["labels_comp_merged"],
            traj["mapped_ref_merged"], traj["mapped_comp_merged"]))

    if "IDF1" in metrics:
        results.update(idf1(
            traj["labels_ref_merged"], traj["labels_comp_merged"],
            traj["mapped_ref_merged"], traj["mapped_comp_merged"]))

    if "MTML" in metrics:
        results.update(mtml(
            traj["labels_ref_merged"], traj["labels_comp_merged"],
            traj["mapped_ref_merged"], traj["mapped_comp_merged"]))

    if "FAF" in metrics:
        results.update(faf(
            traj["labels_comp_merged"], traj["mapped_comp_merged"]))

    return results


def evaluate_sequence(
        res: str,
        gt: str,
        metrics: list = None,
        threads: int = 0,
    ):
    """
    Evaluates a single sequence.

    Args:
        res: The path to the results.
        gt: The path to the ground truth.
        metrics: The metrics to evaluate.
        threads: The number of threads to use. If 0, the number of threads
            is set to the number of available CPUs.

    Returns:
        The results stored in a dictionary.
    """

    print("Evaluate sequence: ", res, " with ground truth: ", gt, end="")
    # Verify all metrics
    if metrics is None:
        metrics = ALL_METRICS

    trajectory_data = True
    segmentation_data = True

    if metrics in [["SEG"], ["CCA"]]:
        trajectory_data = False

    if "SEG" not in metrics:
        segmentation_data = False

    comp_tracks, ref_tracks, traj, segm, _, is_valid = load_data(
        res, gt, trajectory_data, segmentation_data, threads)

    results = calculate_metrics(
        comp_tracks, ref_tracks, traj, segm, metrics, is_valid)

    print("with results: ", results, " done!")

    return results


def evaluate_all(
        res_root: str,
        gt_root: str,
        metrics: list = None,
        threads: int = 0
    ):
    """
    Evaluate all sequences in a directory

    Args:
        res_root: The root directory of the results.
        gt_root: The root directory of the ground truth.
        metrics: The metrics to evaluate.
        threads: The number of threads to use. If 0, the number of threads
            is set to the number of available CPUs.

    Returns:
        The results stored in a dictionary.
    """
    results = []
    ret = parse_directories(res_root, gt_root)
    for res, gt, name in zip(*ret):
        results.append([name, evaluate_sequence(res, gt, metrics, threads)])
    return results


def parse_args():
    """ Parse arguments """
    parser = argparse.ArgumentParser(description='Evaluates CTC-Sequences.')
    parser.add_argument('--res', type=str, required=True)
    parser.add_argument('--gt', type=str, required=True)
    parser.add_argument('-r', '--recursive', action="store_true")
    parser.add_argument('--csv-file', type=str, default=None)
    parser.add_argument('-n', '--num-threads', type=int, default=0)
    parser.add_argument('--valid', action="store_true")
    parser.add_argument('--det', action="store_true")
    parser.add_argument('--seg', action="store_true")
    parser.add_argument('--tra', action="store_true")
    parser.add_argument('--ct', action="store_true")
    parser.add_argument('--tf', action="store_true")
    parser.add_argument('--bc', action="store_true")
    parser.add_argument('--cca', action="store_true")
    parser.add_argument('--mota', action="store_true")
    parser.add_argument('--hota', action="store_true")
    parser.add_argument('--idf1', action="store_true")
    parser.add_argument('--chota', action="store_true")
    parser.add_argument('--mtml', action="store_true")
    parser.add_argument('--faf', action="store_true")
    parser.add_argument('--lnk', action="store_true")
    args = parser.parse_args()
    return args


def main():
    """
    Main function that is called when the script is executed.
    """
    args = parse_args()
    # Prepare metric selection
    metrics = [metric for metric, flag in (
        ("Valid", args.valid),
        ("DET", args.det),
        ("SEG", args.seg),
        ("TRA", args.tra),
        ("CT", args.ct),
        ("TF", args.tf),
        ("BC", args.bc),
        ("CCA", args.cca),
        ("MOTA", args.mota),
        ("HOTA", args.hota),
        ("CHOTA", args.chota),
        ("IDF1", args.idf1),
        ("MTML", args.mtml),
        ("FAF", args.faf),
        ("LNK", args.lnk),
    ) if flag]
    metrics = metrics if metrics else None
    # Evaluate sequence or whole directory
    if args.recursive:
        res = evaluate_all(
            res_root=args.res, gt_root=args.gt, metrics=metrics,
            threads=args.num_threads
        )
    else:
        res = evaluate_sequence(
            res=args.res, gt=args.gt, metrics=metrics, threads=args.num_threads)
    # Visualize and store results
    print_results(res)
    if args.csv_file is not None:
        store_results(args.csv_file, res)


if __name__ == "__main__":
    main()



import os
from os import listdir
from os.path import join, exists, isdir
import numpy as np


def parse_directories(
        input_dir: str,
        gt_dir: str = None,
):
    """
    Parses a directory and searches for challenges and their respective
    result/ground-truth subdirectories.

    Args:
        input_dir: The directory to parse.
        gt_dir: The directory containing the ground truth.

    Returns:
        A tuple of three lists containing the result directories, the
        ground-truth directories and the names of the challenges/sequences.
    """

    # Parse sequences to evaluate
    challenges = [x for x in sorted(listdir(input_dir))
                  if isdir(join(input_dir, x))]
    assert len(challenges) > 0, f"No challenges found in {input_dir}"
    sequence_appendices = ["01", "02"]
    res_dirs, gt_dirs, names = [], [], []
    for challenge in challenges:
        sequences = [
            x[0:2] for x in sorted(listdir(join(input_dir, challenge))) if
            isdir(join(input_dir, challenge, x))
        ]
        for sequence in sequences:
            if sequence in sequence_appendices:
                res_dir = join(input_dir, challenge, sequence + "_RES")
                if res_dir in res_dirs:
                    continue
                res_dirs.append(res_dir)
                if gt_dir is not None:
                    _gt_dir = join(gt_dir, challenge, sequence + "_GT")
                    if not exists(_gt_dir):
                        _gt_dir = None
                    gt_dirs.append(_gt_dir)
                else:
                    gt_dirs.append(None)
                names.append(challenge + "_" + sequence)
    return res_dirs, gt_dirs, names


def read_tracking_file(
        path: str,
):
    """
    Reads a text file representing an acyclic graph for the whole video.
    Every line corresponds to a single track that is encoded by four numbers
    separated by a space:
        L B E P where
        L - a unique label of the track (label of markers, 16-bit positive value)
        B - a zero-based temporal index of the frame in which the track begins
        E - a zero-based temporal index of the frame in which the track ends
        P - label of the parent track (0 is used when no parent is defined)

    Args:
        path: Path to the text file.

    Returns:
        A numpy array of shape (N, 4) where N is the number of tracks.
        Each row represents a track and contains four numbers: L B E P
    """
    if not exists(path):
        return None
    with open(path, "r", encoding="utf-8") as f:
        lines = f.readlines()
    if len(lines) == 0:
        return np.zeros((0, 4))
    seperator = " " if " " in lines[0] else "\t"
    lines = [x.strip().split(seperator) for x in lines]
    lines = [[int(y) for y in x if y != ""] for x in lines]
    return np.array(lines)


def parse_masks(
        directory: str
):
    """
    Reads all frame files in a directory and returns a list of frames.

    Args:
        directory: The directory to read.

    Returns:
        A sorted list of frame paths.
    """
    files = sorted(os.listdir(directory))
    files = [x for x in files if x.endswith(".tif")]
    _files = []
    for x in files:
        if x.count("_") == 3:
            # This is a 3D mask file with slices. Remove the slice number.
            x = "_".join(x.split("_")[0:3]) + ".tif"
        if x not in _files:
            _files.append(join(directory, x))
    files = sorted(_files)
    return files




def tra(
        NS: int,
        FN: int,
        FP: int,
        ED: int,
        EA: int,
        EC: int,
        num_vertices: int,
        num_edges: int,
        **_ignored
):
    """
    Calculate Tracking (TRA) metric.

    According to
        Cell Tracking Accuracy Measurement Based on Comparison of Acyclic
        Oriented Graphs; Matula etal. 2015

    Args:
        NS: Split vertex operations.
        FN: Number of false negatives.
        FP: Number of false positives.
        ED: Number of redundant edge.
        EA: Number of missing edges.
        EC: Number of wrong edge semantics.
        num_vertices: Number of vertices in the graph.
        num_edges: Number of edges in the graph.
        _ignored: Ignored arguments.

    Returns:
        The Tracking metric.
    """
    # Calculate AOGM
    w_ns = 5
    w_fn = 10
    w_fp = 1
    w_ed = 1
    w_ea = 1.5
    w_ec = 1
    AOGM = w_ns * NS + w_fn * FN + w_fp * FP + w_ed * ED + w_ea * EA + w_ec * EC
    # Calculate AOGM_0 (create graph from scratch)
    #   i.e, all vertices and edges are false negatives
    AOGM_0 = w_fn * num_vertices + w_ea * num_edges
    # Calculate DET
    TRA = 1 - min(AOGM, AOGM_0) / AOGM_0
    return float(TRA), AOGM, AOGM_0




import numpy as np

from ctc_metrics.utils.representations import track_confusion_matrix


def hota(
        labels_ref: list,
        labels_comp: list,
        mapped_ref: list,
        mapped_comp: list,
):
    """
    Computes the HOTA metric. As described in the paper,
         "HOTA: A Higher Order Metric for Evaluating Multi-Object Tracking"
           - Luiten et al., {International Journal of Computer Vision 2020

    Args:
        labels_comp: The labels of the computed masks. A list of length equal
            to the number of frames. Each element is a list with the labels of
            the computed masks in the respective frame.
        labels_ref: The labels of the ground truth masks. A list of length
            equal to the number of frames. Each element is a list with the
            labels of the ground truth masks in the respective frame.
        mapped_ref: The matched labels of the ground truth masks. A list of
            length equal to the number of frames. Each element is a list with
            the matched labels of the ground truth masks in the respective
            frame. The elements are in the same order as the corresponding
            elements in mapped_comp.
        mapped_comp: The matched labels of the result masks. A list of length
            equal to the number of frames. Each element is a list with the
            matched labels of the result masks in the respective frame. The
            elements are in the same order as the corresponding elements in
            mapped_ref.

    Returns:
        The HOTA tracks metric.
    """
    # Gather association data
    max_label_ref = int(np.max(np.concatenate(labels_ref)))
    max_label_comp = int(np.max(np.concatenate(labels_comp)))
    track_intersection = track_confusion_matrix(
        labels_ref, labels_comp, mapped_ref, mapped_comp)
    # Calculate Association scores
    hota_score = 0
    for i in range(1, max_label_ref + 1):
        for j in range(1, max_label_comp + 1):
            if track_intersection[i, j] > 0:
                # Calculate the HOTA score
                tpa = track_intersection[i, j]
                fna = np.sum(track_intersection[i, :]) - tpa
                fpa = np.sum(track_intersection[:, j]) - tpa
                a_corr = tpa / (tpa + fna + fpa)
                hota_score += tpa * a_corr
    # Calculate the HOTA score
    tp = track_intersection[1:, 1:].sum()
    fp = track_intersection[0, 1:].sum()
    fn = track_intersection[1:, 0].sum()
    hota_score = np.sqrt(hota_score / (tp + fp + fn))

    res = {
        "HOTA": hota_score,
    }
    return res




def lnk(
        ED: int,
        EA: int,
        EC: int,
        num_edges: int,
        **_ignored
):
    """
    Calculate Linking (LNK) metric.

    According to
        Cell Tracking Accuracy Measurement Based on Comparison of Acyclic
        Oriented Graphs; Matula etal. 2015

    Args:
        ED: Number of redundant edge.
        EA: Number of missing edges.
        EC: Number of wrong edge semantics.
        num_edges: Number of edges in the graph.
        _ignored: Ignored arguments.

    Returns:
        The Linking metric.
    """
    # Calculate AOGM_A
    w_ed = 1
    w_ea = 1.5
    w_ec = 1
    AOGM_A = w_ed * ED + w_ea * EA + w_ec * EC
    # Calculate AOGM_0 (create graph from scratch)
    #   i.e, all vertices and edges are false negatives
    AOGM_0 = w_ea * num_edges
    # Calculate DET
    LNK = 1 - min(AOGM_A, AOGM_0) / AOGM_0
    return float(LNK)




import numpy as np

from ctc_metrics.utils.representations import assign_comp_to_ref


def ct(
        comp_tracks: np.ndarray,
        ref_tracks: np.ndarray,
        labels_ref: list,
        mapped_ref: list,
        mapped_comp: list
):
    """
    Computes the complete tracks metric. As described in the paper,
         "An objective comparison of cell-tracking algorithms."
           - Vladimir Ulman et al., Nature methods 2017

    Args:
        comp_tracks: The result tracks. A (n,4) numpy ndarray with columns:
            - label
            - birth frame
            - end frame
            - parent
        ref_tracks: The ground truth tracks. A (n,4) numpy ndarray with columns:
            - label
            - birth frame
            - end frame
            - parent
        labels_ref: The labels of the ground truth masks. A list of length
            equal to the number of frames. Each element is a list with the
            labels of the ground truth masks in the respective frame.
        mapped_ref: The matched labels of the ground truth masks. A list of
            length equal to the number of frames. Each element is a list with
            the matched labels of the ground truth masks in the respective
            frame. The elements are in the same order as the corresponding
            elements in mapped_comp.
        mapped_comp: The matched labels of the result masks. A list of length
            equal to the number of frames. Each element is a list with the
            matched labels of the result masks in the respective frame. The
            elements are in the same order as the corresponding elements in
            mapped_ref.

    Returns:
        The complete tracks metric.
    """
    track_assignments = assign_comp_to_ref(labels_ref, mapped_ref, mapped_comp)
    correct_tracks = 0
    for k, v in track_assignments.items():
        start_ref = ref_tracks[ref_tracks[:, 0] == k][0, 1]
        end_ref = ref_tracks[ref_tracks[:, 0] == k][0, 2]
        assigned_labels = np.unique(v[~np.isnan(v)])
        if len(assigned_labels) > 1:
            continue
        if assigned_labels[0] == 0:
            continue
        assignee = assigned_labels[0]
        start_comp = comp_tracks[comp_tracks[:, 0] == assignee][0, 1]
        end_comp = comp_tracks[comp_tracks[:, 0] == assignee][0, 2]
        if start_ref == start_comp and end_ref == end_comp:
            correct_tracks += 1
    T_rc = correct_tracks
    T_r = len(ref_tracks)
    T_c = len(comp_tracks)
    return float(2 * T_rc / (T_c + T_r))




import numpy as np

from ctc_metrics.utils.representations import track_confusion_matrix


def mtml(
        labels_ref: list,
        labels_comp: list,
        mapped_ref: list,
        mapped_comp: list
):
    """
    Computes the mostly tracked and mostly lost metric. As described by
    [motchallenge](https://motchallenge.net/).

    Args:
        labels_comp: The labels of the computed masks.
        labels_ref: The labels of the ground truth masks.
        mapped_ref: The matched labels of the ground truth masks.
        mapped_comp: The matched labels of the result masks.

    Returns:
        The mt and ml metrics.
    """
    # Gather association data
    track_intersection = track_confusion_matrix(
        labels_ref, labels_comp, mapped_ref, mapped_comp)
    # Calculate the metrics
    total_ref = np.sum(track_intersection[1:, :], axis=1)
    ratio = np.max(track_intersection[1:, :], axis=1) / np.maximum(total_ref, 1)
    valid_ref = total_ref > 0
    mt = np.sum(ratio[valid_ref] >= 0.8) / np.sum(valid_ref)
    ml = np.sum(ratio[valid_ref] < 0.2) / np.sum(valid_ref)

    res = {
        "MT": mt,
        "ML": ml,
    }
    return res



import numpy as np
from scipy.optimize import linear_sum_assignment

from ctc_metrics.utils.representations import track_confusion_matrix


def get_idf1_stats(
        track_intersection: np.ndarray,
        costs: np.ndarray,
        max_label_ref: int,
        max_label_comp: int
):
    """
    Computes the IDF1 stats.

    Args:
        track_intersection: The track intersection matrix.
        costs: The assignment costs matrix.
        max_label_ref: The maximum label of the reference masks.
        max_label_comp: The maximum label of the computed masks.

    Returns:
        The IDF1 stats.
    """
    # Get optimal assignment
    row_ind, col_ind = linear_sum_assignment(costs)
    # Accumulate trackwise number of matches
    all_gt = np.sum(track_intersection[1:], axis=1)
    all_comp = np.sum(track_intersection[:, 1:], axis=0)
    # Compute IDFN, IDFP
    IDFN = all_gt.sum()
    IDFP = all_comp.sum()
    for i, j in zip(row_ind, col_ind):
        if i < max_label_ref and j < max_label_comp:
            IDFN -= track_intersection[i + 1, j + 1]
            IDFP -= track_intersection[i + 1, j + 1]

    assert (track_intersection[1:, :].sum() - IDFN ==
            track_intersection[:, 1:].sum() - IDFP)
    # Compute IDTP
    IDTP = track_intersection[1:, :].sum() - IDFN
    # Compute IDF1, IDP, IDR
    IDP = IDTP / (IDTP + IDFP)
    IDR = IDTP / (IDTP + IDFN)
    IDF1 = 2 * IDTP / (2 * IDTP + IDFP + IDFN)

    return {
        "IDF1": IDF1,
        "IDP": IDP,
        "IDR": IDR,
        "IDTP": IDTP,
        "IDFP": IDFP,
        "IDFN": IDFN
    }

def idf1(
        labels_ref: list,
        labels_comp: list,
        mapped_ref: list,
        mapped_comp: list
):
    """
    Computes the IDF1 metric. As described in the paper,
         "Performance Measures and a Data Set for
          Multi-Target, Multi-Camera Tracking"
           - Ristani et al., ECCV  2016

    Args:
        labels_comp: The labels of the computed masks. A list of length equal
            to the number of frames. Each element is a list with the labels of
            the computed masks in the respective frame.
        labels_ref: The labels of the ground truth masks. A list of length
            equal to the number of frames. Each element is a list with the
            labels of the ground truth masks in the respective frame.
        mapped_ref: The matched labels of the ground truth masks. A list of
            length equal to the number of frames. Each element is a list with
            the matched labels of the ground truth masks in the respective
            frame. The elements are in the same order as the corresponding
            elements in mapped_comp.
        mapped_comp: The matched labels of the result masks. A list of length
            equal to the number of frames. Each element is a list with the
            matched labels of the result masks in the respective frame. The
            elements are in the same order as the corresponding elements in
            mapped_ref.

    Returns:
        The idf1 tracks metric.
    """
    # Calculate the cop-ref track intersections
    max_label_ref = int(np.max(np.concatenate(labels_ref)))
    max_label_comp = int(np.max(np.concatenate(labels_comp)))
    track_intersection = track_confusion_matrix(
        labels_ref, labels_comp, mapped_ref, mapped_comp
    )
    # Create assignment costs matrx with default of inf costs + Line for the
    #  dummy node (no association)
    total_ids = max_label_ref + max_label_comp
    costs = np.ones((total_ids, total_ids)) * np.inf
    # Fill in the costs of intersection between  tracks
    subcosts = np.zeros_like(track_intersection[1:, 1:])
    subcosts[track_intersection[1:, 1:] == 0] = np.inf
    costs[:max_label_ref, :max_label_comp] = subcosts
    costs[:max_label_ref, :max_label_comp] -= 2 * track_intersection[1:, 1:]
    costs[:max_label_ref, :max_label_comp] += np.sum(
        track_intersection[1:, :], axis=1)[:, None]
    costs[:max_label_ref, :max_label_comp] += np.sum(
        track_intersection[:, 1:], axis=0)[None, :]
    # Set the assignment costs to the dummy nodes
    costs[max_label_ref:, max_label_comp:] = 0
    for i, c in enumerate(np.sum(track_intersection[1:, :], axis=1)):
        costs[i, max_label_comp + i] = c
    for j, c in enumerate(np.sum(track_intersection[:, 1:], axis=0)):
        costs[max_label_ref + j, j] = c

    return get_idf1_stats(
        track_intersection, costs, max_label_ref, max_label_comp
    )



from setuptools import setup, find_packages

setup(
    name="py-ctcmetrics",
    version="1.2.2",
    packages=find_packages(),
    install_requires=[
        "numpy",
        "scikit-learn",
        "scipy",
        "tifffile",
        "imagecodecs",
        "pandas"
    ],
    extras_require={
        "viz": ["opencv-python"],
    },
    author="Timo Kaiser",
    author_email="kaiser@tnt.uni-hannover.de",
    description="Metrics for Cell Tracking Challenges",
    long_description="Metrics for Cell Tracking Challenges",
    entry_points={
        'console_scripts': [
            'ctc_evaluate = ctc_metrics.scripts.evaluate:main',
            'ctc_validate = ctc_metrics.scripts.validate:main',
            'ctc_noise = ctc_metrics.scripts.noise:main',
            'ctc_visualize = ctc_metrics.scripts.visualize:main',
            # Add more scripts here if needed
        ],
    },
)



import numpy as np

from ctc_metrics.utils.representations import track_confusion_matrix


def cluster_clique(
        tracks: np.ndarray
):
    """
    Clusters the tracks into cliques. A clique is a set of tracks that are
    connected by parent-child relationships.

    Args:
        tracks: The result tracks.

    Returns:
        The cliques. A dictionary with the track id as key and the clique as
        value.
    """
    # Cluster ref cliques
    track_id = np.unique(tracks[:, 0]).astype(int)
    track_parents = np.asarray(
        [tracks[tracks[:, 0] == x][0, 3] for x in track_id]).astype(int)
    cliques = {x: [] for x in track_id}

    # forward
    for track in track_id:
        clique = np.asarray([track])
        while True:
            is_child = np.isin(track_parents, clique)
            is_not_in_clique = np.isin(track_id, clique, invert=True)
            new_children = track_id[is_child & is_not_in_clique]
            if len(new_children) == 0:
                break
            clique = np.concatenate((clique, new_children))
        cliques[track] = clique

    # backward
    for track in track_id:
        clique = np.zeros(0)
        for ancestor in track_id:
            if track in cliques[ancestor]:
                clique = np.concatenate((clique, [ancestor]))
        cliques[track] = np.concatenate((cliques[track], clique))

    for track in track_id:
        cliques[track] = np.unique(cliques[track]).astype(int)

    return cliques


def chota(
        ref_tracks: np.ndarray,
        comp_tracks: np.ndarray,
        labels_ref: list,
        labels_comp: list,
        mapped_ref: list,
        mapped_comp: list
):
    """


    Args:
        ref_tracks: The ground truth tracks. A (n,4) numpy ndarray with columns:
            - label
            - birth frame
            - end frame
            - parent
        comp_tracks: The computed tracks. A (n,4) numpy ndarray with columns:
            - label
            - birth frame
            - end frame
            - parent
        labels_comp: The labels of the computed masks. A list of length equal
            to the number of frames. Each element is a list with the labels of
            the computed masks in the respective frame.
        labels_ref: The labels of the ground truth masks. A list of length
            equal to the number of frames. Each element is a list with the
            labels of the ground truth masks in the respective frame.
        mapped_ref: The matched labels of the ground truth masks. A list of
            length equal to the number of frames. Each element is a list with
            the matched labels of the ground truth masks in the respective
            frame. The elements are in the same order as the corresponding
            elements in mapped_comp.
        mapped_comp: The matched labels of the result masks. A list of length
            equal to the number of frames. Each element is a list with the
            matched labels of the result masks in the respective frame. The
            elements are in the same order as the corresponding elements in
            mapped_ref.

    Returns:
        The CHOTA tracks metric.
    """
    # Gather association data
    cliques_ref = cluster_clique(ref_tracks)
    cliques_comp = cluster_clique(comp_tracks)
    track_intersection = track_confusion_matrix(
        labels_ref, labels_comp, mapped_ref, mapped_comp)
    # Calculate Association scores
    chota_score = 0
    for i in range(1, int(np.max(np.concatenate(labels_ref))) + 1):
        for j in range(1, int(np.max(np.concatenate(labels_comp))) + 1):
            if track_intersection[i, j] > 0:
                cliques_ref_i = cliques_ref[i]
                cliques_comp_j = cliques_comp[j]
                roi1 = np.zeros_like(track_intersection, dtype=bool)
                roi2 = np.zeros_like(track_intersection, dtype=bool)
                roi1[cliques_ref_i, :] = True
                roi2[:, cliques_comp_j] = True
                # Calculate the HOTA score
                tpa = np.sum(track_intersection[roi1 & roi2])
                fna = np.sum(track_intersection[cliques_ref_i, :]) - tpa
                fpa = np.sum(track_intersection[:, cliques_comp_j]) - tpa
                # Reweight and add to hota score
                num_pixels = track_intersection[i, j]
                a_corr = tpa / (tpa + fna + fpa)
                chota_score += num_pixels * a_corr
    # Calculate the finalCHOTA score
    tp = track_intersection[1:, 1:].sum()
    fp = track_intersection[0, 1:].sum()
    fn = track_intersection[1:, 0].sum()
    chota_score = np.sqrt(chota_score / (tp + fp + fn))
    return {"CHOTA": chota_score}





import numpy as np


def get_ids_that_ends_with_split(
        tracks: np.ndarray
):
    """
    Extracts the ids of tracks that end with a cell split.

    Args:
        tracks: The tracks to check. A numpy nd array with columns:
            - label
            - birth frame
            - end frame
            - parent

    Returns:
        The ids of tracks that end with a cell split stored in a numpy.ndarray.
    """
    parents, counts = np.unique(tracks[:, 3], return_counts=True)
    counts = counts[parents > 0]
    parents = parents[parents > 0]
    ends_with_split = parents[counts > 1]
    return ends_with_split


def calculate_f1_score(
        tp: int,
        fp: int,
        fn: int
):
    """
    Calculates the f1 score.

    Args:
        tp: The number of true positives.
        fp: The number of false positives.
        fn: The number of false negatives.

    Returns:
        The f1 score.
    """
    precision = tp / max((tp + fp), 1)
    recall = tp / max((tp + fn), 1)
    f1_score = 2 * (precision * recall) / max((precision + recall), 0.0001)
    return f1_score


def is_matching(
        id_comp: int,
        id_ref: int,
        mapped_ref: list,
        mapped_comp: list,
        ref_children: np.ndarray,
        comp_children: np.ndarray,
        tr: int,
        tc: int
):
    """
    Checks if the reference and the computed track match.

    Args:
        id_comp: The computed track id.
        id_ref: The reference track id.
        mapped_ref: The matched labels of the ground truth masks.
        mapped_comp: The matched labels of the result masks.
        ref_children: The children ids of the reference track.
        comp_children: The children ids of the computed track.
        tr: The frame of the reference track end.
        tc: The frame of the computed track end.

    Returns:
        True if the reference and the computed track match, False otherwise.
    """
    # Check if the number of children is the same
    if len(ref_children) != len(comp_children):
        return False
    # Compare parents
    t1, t2 = min(tr, tc), max(tr, tc)
    mr, mc = mapped_ref[t1], mapped_comp[t1]
    if np.sum(mc == id_comp) < 1 or np.sum(mr == id_ref) != 1:
        return False
    ind = np.argwhere(mr == id_ref).squeeze()
    if mc[ind] != id_comp:
        return False
    # Compare children
    mr, mc = np.asarray(mapped_ref[t2 + 1]), np.asarray(mapped_comp[t2 + 1])
    if not np.all(np.isin(comp_children, mc)):
        return False
    if not np.all(np.isin(mr[np.isin(mc, comp_children)], ref_children)):
        return False
    return True


def bc(
        comp_tracks: np.ndarray,
        ref_tracks: np.ndarray,
        mapped_ref: list,
        mapped_comp: list,
        i: int
):
    """
    Computes the branching correctness metric. As described in the paper,
         "An objective comparison of cell-tracking algorithms."
           - Vladimir Ulman et al., Nature methods 2017

    Args:
        comp_tracks: The result tracks. A (n,4) numpy ndarray with columns:
            - label
            - birth frame
            - end frame
            - parent
        ref_tracks: The ground truth tracks. A (n,4) numpy ndarray with columns:
            - label
            - birth frame
            - end frame
            - parent
        mapped_ref: The matched labels of the ground truth masks. A list of
            length equal to the number of frames. Each element is a list with
            the matched labels of the ground truth masks in the respective
            frame. The elements are in the same order as the corresponding
            elements in mapped_comp.
        mapped_comp: The matched labels of the result masks. A list of length
            equal to the number of frames. Each element is a list with the
            matched labels of the result masks in the respective frame. The
            elements are in the same order as the corresponding elements in
            mapped_ref.
        i: The maximal allowed error in frames.

    Returns:
        The branching correctness metric.
    """

    # Extract relevant tracks with children in reference
    ends_with_split_ref = get_ids_that_ends_with_split(ref_tracks)
    t_ref = np.array([ref_tracks[ref_tracks[:, 0] == ref][0, 2]
                      for ref in ends_with_split_ref])
    if len(ends_with_split_ref) == 0:
        return None
    # Extract relevant tracks with children in computed result
    ends_with_split_comp = get_ids_that_ends_with_split(comp_tracks)
    t_comp = np.asarray([comp_tracks[comp_tracks[:, 0] == comp][0, 2]
                         for comp in ends_with_split_comp])
    if len(ends_with_split_comp) == 0:
        return 0
    # Find all matches between reference and computed branching events (mitosis)
    matches = []
    for comp, tc in zip(ends_with_split_comp, t_comp):
        # Find potential matches
        pot_matches = np.abs(t_ref - tc) <= i
        if len(pot_matches) == 0:
            continue
        comp_children = comp_tracks[comp_tracks[:, 3] == comp][:, 0]
        # Evaluate potential matches
        for ref, tr in zip(
                ends_with_split_ref[pot_matches],
                t_ref[pot_matches]
        ):
            ref_children = ref_tracks[ref_tracks[:, 3] == ref][:, 0]
            if is_matching(
                    comp, ref, mapped_ref, mapped_comp, ref_children,
                    comp_children, tr, tc
            ):
                matches.append((ref, comp))
    # Calculate BC(i)
    return calculate_f1_score(
        len(matches),
        len(ends_with_split_comp) - len(matches),
        len(ends_with_split_ref) - len(matches))
